﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using WiimoteLib;

namespace WPF_Wii_Binding_Library
{
    public class MouseSource : IInputSource
    {
        private int currentPair = 0;
        private float pairDistance = 0.015f;
        private float mouseX = 0;
        private float mouseY = 0;
        private bool[] fingerVisible = new bool[4] {true, true, true, true};
        private IRSensor[] state;
        
        public MouseSource() {
            state = new IRSensor[4] { new IRSensor(), new IRSensor(), new IRSensor(), new IRSensor() };
            state[0].Found = true;
            state[1].Found = true;
            state[2].Found = true;
            state[3].Found = true;
        }

#region IInputSource Members
        public event InputChangedEventHandler InputChanged;

        WiimoteLib.IRSensor[] IInputSource.GetData() {
            return state;
        }
#endregion

#region Public Methods
        public void InjectMouseMove(double x, double y) {
            mouseX = (float)x;
            mouseY = (float)y;
            Update();
        }

        public void InjectRightClick() {
            currentPair = 1 - currentPair;
            Update();
        }

        public void InjectMouseWheel(int delta) {
            pairDistance += delta > 0 ? 0.005f : -0.005f;
            Update();
        }

        public void InjectKey(System.Windows.Input.Key key) {
            if (key == System.Windows.Input.Key.D1) fingerVisible[0] = !fingerVisible[0];
            if (key == System.Windows.Input.Key.D2) fingerVisible[1] = !fingerVisible[1];
            if (key == System.Windows.Input.Key.D3) fingerVisible[2] = !fingerVisible[2];
            if (key == System.Windows.Input.Key.D4) fingerVisible[3] = !fingerVisible[3];
            Update();
        }
#endregion

#region Private Methods
        private void Update() {
            if (currentPair == 0) {
                state[0].Position = new System.Drawing.PointF(mouseX, mouseY - pairDistance);
                state[1].Position = new System.Drawing.PointF(mouseX, mouseY + pairDistance);
            } else {
                state[2].Position = new System.Drawing.PointF(mouseX, mouseY - pairDistance);
                state[3].Position = new System.Drawing.PointF(mouseX, mouseY + pairDistance);
            }
            state[0].Found = fingerVisible[0];
            state[1].Found = fingerVisible[1];
            state[2].Found = fingerVisible[2];
            state[3].Found = fingerVisible[3];
            if (InputChanged != null) InputChanged();
        }
#endregion

    }
}
