﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using WiimoteLib;
using System.Windows;
using System.Diagnostics;

namespace WPF_Wii_Binding_Library
{
    using History = List<List<Vector>>;

    public class Pairing
    {
#region Private members
        IInputSource source;        //Input
        FingerBinding binding;      //Output

        //Internal state
        History fingerHistory = new History();
        FingerPoint[] fingerpoints = new FingerPoint[4] { new FingerPoint(), new FingerPoint(), new FingerPoint(), new FingerPoint() };
        FingerPoint[] tmpfingerpoints = new FingerPoint[4];
        int fingerpointCount, fingerpointOldCount, tmpFingerpointCount;
        FingerPair[] fingerPairs = new FingerPair[2] { new FingerPair(), new FingerPair() };
        double maximumClickDistance = 0.25;
        int lastFrameS = 0;
        int AvgFPS = 100;   //input speed of the Wii
        int numFrames = 0;
        double inputScaling = 0.8;
#endregion

        public Pairing(IInputSource source, FingerBinding binding) {
            this.source = source;
            this.binding = binding;
            source.InputChanged += Source_InputChanged;
        }

#region Private Methods
        //Event handler of source.InputChanged
        private void Source_InputChanged() {
            DoTrackPoints(source.GetData());
            if (DateTime.Now.Second != lastFrameS) {
                AvgFPS = numFrames;
                numFrames = 0;
                lastFrameS = DateTime.Now.Second;
            }
            DoPairing(source.GetData(), AvgFPS);
            ShowPairs();
            numFrames++;
        }

        private void ShowPoints()
        {
            binding.P1F1Position = ToPositionFromFinger(0);
            binding.P1F1Visible = true;
            binding.P1F2Position = ToPositionFromFinger(1);
            binding.P1F2Visible = true;
            binding.P2F1Position = ToPositionFromFinger(2);
            binding.P2F1Visible = true;
            binding.P2F2Position = ToPositionFromFinger(3);
            binding.P2F2Visible = true;
        }

        private void ShowPairs()
        {
            //update the finger objects
            binding.P1F1Visible = fingerPairs[0].active > 0;
            binding.P1F1Size = fingerPairs[0].pinched ? 60 : 30;
            binding.P1F1Position = ToPositionFromFinger(fingerPairs[0].fingerids[0]);

            binding.P1F2Visible = fingerPairs[0].active > 1;
            binding.P1F2Position = ToPositionFromFinger(fingerPairs[0].fingerids[1]);

            binding.P2F1Visible = fingerPairs[1].active > 0;
            binding.P2F1Size = fingerPairs[1].pinched ? 60 : 30;
            binding.P2F1Position = ToPositionFromFinger(fingerPairs[1].fingerids[0]);

            binding.P2F2Visible = fingerPairs[1].active > 1;
            binding.P2F2Position = ToPositionFromFinger(fingerPairs[1].fingerids[1]);
        }

        private void DoTrackPoints(IRSensor[] points) {
            tmpFingerpointCount = points.Count((point) => point.Found);
            for (int i = 0; i < 4; i++) tmpfingerpoints[i] = null;
	        for (int i = 0, count=0; i < 4; i++) {
                if (points[i].Found)
                {
                    tmpfingerpoints[count] = new FingerPoint(points[i].Position, count, inputScaling, binding);
                    count++;
                }
 	        }

            int chosen;
            FingerPoint[] newPoints = new FingerPoint[4];
            bool[] point_placed = new bool[4];
            if (fingerpointCount != tmpFingerpointCount)
            {
                if (fingerpointCount < tmpFingerpointCount) {	//points added
			        //loop over the old points, and remember which new points we used
                    bool[] id_used = new bool[4];

                    for (int i = 0; i < fingerpointCount; i++)
                    {
				        //find the new corresponding point
                        chosen = FindClosestPoint(tmpfingerpoints, fingerpoints[i], ref point_placed);

				        newPoints[i] = new FingerPoint(tmpfingerpoints[chosen].point, fingerpoints[i].id);
				        point_placed[chosen] = true;
				        id_used[fingerpoints[i].id] = true;
			        }
			        //now loop over the new points that aren't placed yet and give them an id
			        for (int i=0; i<tmpFingerpointCount; i++) {
				        if (!point_placed[i]) {
					        for (int j=0; j<4; j++) {
						        if (!id_used[j]) {
                                    newPoints[fingerpointCount] = new FingerPoint(tmpfingerpoints[i].point, j);
                                    id_used[j] = true;
							        fingerpointCount++;
							        break;
						        }
					        }
				        }
			        }
		        } else {							//points removed
                    //loop over the new points
                    for (int i = 0; i < tmpFingerpointCount; i++)
                    {
                        //find the new corresponding point
                        chosen = FindClosestPoint(fingerpoints, tmpfingerpoints[i], ref point_placed);
                        newPoints[i] = new FingerPoint(tmpfingerpoints[i].point, fingerpoints[chosen].id);
                    }
                }
		        fingerpointCount = tmpFingerpointCount;
	        } else {
                //no points have really changed, just find the closest ones
                for (int i = 0; i < tmpFingerpointCount; i++)
                {
                    chosen = FindClosestPoint(fingerpoints, tmpfingerpoints[i], ref point_placed);
                    newPoints[chosen] = new FingerPoint(tmpfingerpoints[i].point, fingerpoints[chosen].id);
                }
            }
            fingerpoints = newPoints;
        }

        private void DoPairing(IRSensor[] points, double FPS) {
	        if (fingerpointOldCount != fingerpointCount) {
		        fingerHistory.Clear();

		        int fPC=fingerpointOldCount;
		        if (fPC < fingerpointCount) {
			        fPC++;
		        } else {
			        fPC--;
		        }
		        do {
			        if (fingerpointOldCount <= fPC) {	//points added
				        double min, d;
				        int chosen = 0, pair = 0;
				        switch (fPC) {
					        case 1:
						        fingerPairs[0].fingerids[0] = findIdNotUsed();
						        fingerPairs[0].active = 1;
						        fingerPairs[1].active = 0;
						        break;
					        case 2:
						        pair = (fingerPairs[0].active < 1 ? 0 : 1);
						        fingerPairs[pair].fingerids[0] = findIdNotUsed();;
						        fingerPairs[pair].active = 1;
						        break;
					        case 3:
						        //find the closest two points
						        if (fingerPairs[0].active == 2 || fingerPairs[1].active == 2) {
							        pair = fingerPairs[1].active < 2 ? 1 : 0;
							        fingerPairs[pair].fingerids[0] = findIdNotUsed();
							        fingerPairs[pair].fingerids[1] = -1;
							        fingerPairs[pair].active = 1;
						        } else {
							        min = 2;
							        for (int i=0; i<3; i++) {
								        d = Dist(fingerpoints[i], fingerpoints[(i+1)%3], 1.8);	//favor the y direction bigtime
								        if (d < min) {
									        chosen = i;
									        min = d;
								        }
							        }
							        //try to make as less changes to the pairs as possible
							        switch (chosen) {
								        case 0:	//need to do reordering, since point 0 and 1 need to form a pair
									        pair = fingerPairs[0].active < 2 ? 1 : 0;
									        fingerPairs[pair].fingerids[0] = fingerpoints[0].id;
									        fingerPairs[pair].fingerids[1] = fingerpoints[1].id;
									        fingerPairs[pair].active = 2;
									        fingerPairs[1 - pair].fingerids[0] = fingerpoints[2].id;
									        fingerPairs[1 - pair].fingerids[1] = -1;
									        fingerPairs[1 - pair].active = 1;
									        break;
								        case 1:	//point 1 and 2 need to form a pair, add point 2 to group of point 1
									        pair = fingerPairs[1].fingerids[0] == fingerpoints[1].id ? 1 : 0;
									        fingerPairs[pair].fingerids[1] = fingerpoints[2].id;
									        fingerPairs[pair].active = 2;
									        fingerPairs[1 - pair].active = 1;
									        break;
								        case 2: //point 2 and 0 need to form a pair, add point 2 to group of point 0
									        pair = fingerPairs[1].fingerids[0] == fingerpoints[0].id ? 1 : 0;
									        fingerPairs[pair].fingerids[1] = fingerpoints[2].id;
									        fingerPairs[pair].active = 2;
									        fingerPairs[1 - pair].active = 1;
									        break;
							        }
						        }
                                if (fingerPairs[pair].active == 2) {
                                    fingerPairs[pair].distance = Dist(fingerpoints[FindFingerById(fingerPairs[pair].fingerids[0])], fingerpoints[FindFingerById(fingerPairs[pair].fingerids[1])], 1.0);
                                } else {
                                    fingerPairs[pair].distance = 0;
                                }
						        FingerPairLost(pair);
						        break;
					        case 4:
						        pair = (fingerPairs[0].active < 2 ? 0 : 1);
						        fingerPairs[pair].fingerids[1] = findIdNotUsed();
						        fingerPairs[pair].distance = Dist(fingerpoints[FindFingerById(fingerPairs[pair].fingerids[0])], fingerpoints[FindFingerById(fingerPairs[pair].fingerids[1])], 1.0);
                                fingerPairs[pair].active = 2;
						        FingerPairLost(pair);
						        break;
				        }
				        fPC++;
				        fingerpointOldCount++;
			        } else {										//points removed
				        for (int i=0; i<2; i++) {
					        for (int j=0; j<2; j++) {
						        if (fingerPairs[i].fingerids[j] != -1) {
							        bool found = false;
                                    for (int k = 0; k < fingerpoints.Count((point) => point != null); k++) {
								        if (fingerpoints[k].id == fingerPairs[i].fingerids[j]) {
									        found = true;
									        break;
								        }
							        }
							        if (!found) {
								        if (j == 0 && fingerPairs[i].active == 2) {
									        //put the still active point in point 0 of the pair, easier when adding points
									        fingerPairs[i].fingerids[0] = fingerPairs[i].fingerids[1];
									        fingerPairs[i].fingerids[1] = -1;
								        } else {
									        fingerPairs[i].fingerids[j] = -1;
								        }
								        if (fingerPairs[i].active == 2) {
									        if (fingerPairs[i].distance < maximumClickDistance) {
										        FingerPairCreated(i);
									        }
								        } else if (fingerPairs[i].active == 1) {
									        FingerPairLost(i);
								        }
								        fingerPairs[i].active--;
							        }
						        }
					        }
				        }
				        fPC--;
				        fingerpointOldCount--;
			        }
		        } while (fingerpointOldCount != fingerpointCount);
	        } else {
		        //add the points to the history
		        if (fingerpointCount == 4) {
                    List<Vector> tmp = new List<Vector>();
			        int i0 = FindFingerById(0), i1 = FindFingerById(1), i2 = FindFingerById(2), i3 = FindFingerById(3);
                    tmp.Add(new Vector(fingerpoints[i0].point.X, fingerpoints[i0].point.Y));
                    tmp.Add(new Vector(fingerpoints[i1].point.X, fingerpoints[i1].point.Y));
                    tmp.Add(new Vector(fingerpoints[i2].point.X, fingerpoints[i2].point.Y));
                    tmp.Add(new Vector(fingerpoints[i3].point.X, fingerpoints[i3].point.Y));
			        fingerHistory.Add(tmp);
                    if (fingerHistory.Count > FPS * 0.3) { //wait 0.2 seconds
                        Vector[] directions = new Vector[4];
				        for (int i=0; i<4; i++) {
					        directions[i] = fingerHistory[0][i] - fingerHistory[fingerHistory.Count - 1][i];
				        }
				        if (directions[0].Length > 0.05 || directions[1].Length > 0.05 || directions[2].Length > 0.05 || directions[3].Length > 0.05) {
					        bool[] contraMovement = new bool[2];
					        contraMovement[0] = contraMovement[1] = false;
					        for (int i=0; i<2; i++) {
						        int j0 = fingerPairs[i].fingerids[0], j1 = fingerPairs[i].fingerids[1];
                                if (angleDegree(directions[j0], directions[j1]) > 120) {
							        contraMovement[i] = true;
						        } else {
							        contraMovement[i] = false;
						        }
					        }
					        if (contraMovement[0] && contraMovement[1]) {
						        //swap the first finger of the first pair with one of the other pair => the one with the lowest dot product
						        int i00 = fingerPairs[0].fingerids[0];
						        int i10 = fingerPairs[1].fingerids[0], i11 = fingerPairs[1].fingerids[1];
						        if (dot(directions[i00], directions[i10]) < dot(directions[i00], directions[i11])) {
							        //swap with i10
							        fingerPairs[1].fingerids[0] = i00;
							        fingerPairs[0].fingerids[0] = i10;
						        } else {
							        //swap with i11
							        fingerPairs[1].fingerids[1] = i00;
							        fingerPairs[0].fingerids[0] = i11;
						        }
						        fingerHistory.Clear();
					        }
				        }
                        while (fingerHistory.Count > FPS * 0.33) {
					        fingerHistory.RemoveAt(fingerHistory.Count-1);
				        }
			        }
		        }
	        }
	        for (int i=0; i<2; i++) {
		        if (fingerPairs[i].active == 1) {
			        FingerPairMoved(i);
		        }
		        if (fingerPairs[i].active == 2) {
			        fingerPairs[i].distance = Dist(fingerpoints[FindFingerById(fingerPairs[i].fingerids[0])], fingerpoints[FindFingerById(fingerPairs[i].fingerids[1])], 1.0);
		        }
	        }
	        fingerpointOldCount = fingerpointCount;
        }
#endregion

#region Supporting functions

        private int FindClosestPoint(FingerPoint[] pointsToSearch, FingerPoint referencePoint, ref bool[] placed)
        {
            double minDist = double.MaxValue, d;
            int chosen = -1;

            for (int i = 0; i < pointsToSearch.Count((point) => point != null); i++)
            {
                if (!placed[i] && (d = Dist(pointsToSearch[i], referencePoint, 1.0)) < minDist)
                {
                    minDist = d;
                    chosen = i;
                }
            }
            placed[chosen] = true;
            return chosen;
        }

        private int FindFingerById(int id) {
            return FindFingerById(id, false);
        }

        private int FindFingerById(int id, bool correct) {
	        for (int i=0; i<fingerpointCount; i++){ 
		        if (fingerpoints[i].id == id) {
			        return i;
		         }
	        }
            return (correct ? -1 : 0);
        }

        System.Windows.Point ToPositionFromFinger(int id) {
            int index = FindFingerById(id, true);
            if (index == -1) return new Point(-30, -30);
            return new Point(fingerpoints[index].point.X * binding.CanvasWidth, fingerpoints[index].point.Y * binding.CanvasHeight);
        }

        void FingerPairCreated(int pair) {
            fingerPairs[pair].pinched = true;
            //TODO: Do something here?
        }

        void FingerPairLost(int pair) {
            fingerPairs[pair].pinched = false;
            //TODO: Do something here?
        }

        void FingerPairMoved(int pair) {
            //TODO: Do something here?
        }

        private double Dist(FingerPoint a, FingerPoint b, double penalty_x)
        {
            double dx = penalty_x * (b.point.X - a.point.X), dy = b.point.Y - a.point.Y;
            return Math.Sqrt(dx * dx + dy * dy);
        }

        private int findIdNotUsed() {
            for (int i = 0; i < fingerpoints.Count((point) => point != null); i++) {
		        bool used = false;
		        for (int j=0; j<2; j++) {
			        used = used || fingerpoints[i].id == fingerPairs[j].fingerids[0] || fingerpoints[i].id == fingerPairs[j].fingerids[1];
		        }
		        if (!used) {
			        return fingerpoints[i].id;
		        }
	        }
	        return 0;
        }

        private double dot(Vector a, Vector b) {
            return a.X * b.X + a.Y * b.Y;
        }

        private double angleDegree(Vector a, Vector b) {
            return Math.Acos(dot(a, b) / (a.Length * b.Length)) / (2 * Math.PI) * 360;
        }

#endregion

    }
}